# Copyright © 2020, Joseph Berry, Rico Tabor (opendrop.dev@gmail.com)
# OpenDrop is released under the GNU GPL License. You are free to
# modify and distribute the code, but always under the same license
#
# If you use this software in your research, please cite the following
# journal articles:
#
# J. D. Berry, M. J. Neeson, R. R. Dagastine, D. Y. C. Chan and
# R. F. Tabor, Measurement of surface and interfacial tension using
# pendant drop tensiometry. Journal of Colloid and Interface Science 454
# (2015) 226–237. https://doi.org/10.1016/j.jcis.2015.05.012
#
# E. Huang, T. Denning, A. Skoufis, J. Qi, R. R. Dagastine, R. F. Tabor
# and J. D. Berry, OpenDrop: Open-source software for pendant drop
# tensiometry & contact angle measurements, submitted to the Journal of
# Open Source Software
#
# These citations help us not only to understand who is using and
# developing OpenDrop, and for what purpose, but also to justify
# continued development of this code and other open source resources.
#
# OpenDrop is distributed WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.  You
# should have received a copy of the GNU General Public License along
# with this software.  If not, see <https://www.gnu.org/licenses/>.

from utils.misc import rotation_mat2d
from modules.ift.younglaplace.shape import YoungLaplaceShape

from typing import Sequence, Tuple, NamedTuple, Optional
from enum import IntEnum, auto
import math
import numpy as np
import scipy.optimize


__all__ = (
    "YoungLaplaceFitResult",
    "young_laplace_fit",
)

DELTA_TOL = 1.0e-8
GRADIENT_TOL = 1.0e-8
OBJECTIVE_TOL = 1.0e-8
MAX_STEPS = 50
# Math constants.
PI = math.pi
NAN = math.nan


class YoungLaplaceFitResult(NamedTuple):
    bond: float
    radius: float
    apex_x: float
    apex_y: float
    rotation: float

    objective: float
    residuals: np.ndarray
    closest: np.ndarray
    arclengths: np.ndarray

    volume: float
    surface_area: float


def young_laplace_fit(data: Tuple[np.ndarray, np.ndarray], verbose: bool = False):
    model = YoungLaplaceModel(data)

    def fun(params: Sequence[float], model: YoungLaplaceModel) -> np.ndarray:
        model.set_params(params)
        return model.residuals

    def jac(params: Sequence[float], model: YoungLaplaceModel) -> np.ndarray:
        model.set_params(params)
        return model.jac

    initial_params = young_laplace_guess(data)
    if initial_params is None:
        raise ValueError("Parameter estimatation failed for this data set")

    model.set_params(initial_params)

    optimize_result = scipy.optimize.least_squares(
        fun,
        model.params,
        jac,
        args=(model,),
        x_scale="jac",
        method="lm",
        ftol=OBJECTIVE_TOL,
        xtol=DELTA_TOL,
        gtol=GRADIENT_TOL,
        verbose=2 if verbose else 0,
        max_nfev=MAX_STEPS,
    )

    # Update model parameters to final result.
    model.set_params(optimize_result.x)

    result = YoungLaplaceFitResult(
        bond=model.params[YoungLaplaceParam.BOND],
        radius=model.params[YoungLaplaceParam.RADIUS],
        apex_x=model.params[YoungLaplaceParam.APEX_X],
        apex_y=model.params[YoungLaplaceParam.APEX_Y],
        rotation=model.params[YoungLaplaceParam.ROTATION],
        objective=(model.residuals**2).sum() / model.dof,
        residuals=model.residuals,
        closest=model.closest,
        arclengths=model.arclengths,
        volume=model.volume,
        surface_area=model.surface_area,
    )

    return result


def young_laplace_guess(data: Tuple[np.ndarray, np.ndarray]) -> Optional[tuple]:
    from modules.ift.pendant import find_pendant_apex

    params = np.empty(len(YoungLaplaceParam))

    ans = find_pendant_apex(data)
    if ans is None:
        return None

    apex, radius, rotation = ans

    r, z = rotation_mat2d(-rotation) @ (data - np.reshape(apex, (2, 1)))
    bond = _bond_selected_plane(r, z, radius)

    params[YoungLaplaceParam.BOND] = bond
    params[YoungLaplaceParam.RADIUS] = radius
    params[YoungLaplaceParam.APEX_X] = apex.x
    params[YoungLaplaceParam.APEX_Y] = apex.y
    params[YoungLaplaceParam.ROTATION] = rotation

    return params


def _bond_selected_plane(r: np.ndarray, z: np.ndarray, radius: float) -> float:
    """Estimate Bond number by method of selected plane."""
    z_ix = np.argsort(z)
    if np.searchsorted(z, 2.0 * radius, sorter=z_ix) < len(z):
        lower, upper = np.searchsorted(z, [1.95 * radius, 2.05 * radius], sorter=z_ix)
        radii = np.abs(r[z_ix][lower : upper + 1])
        x = radii.mean() / radius
        bond = max(0.10, 0.1756 * x**2 + 0.5234 * x**3 - 0.2563 * x**4)
    else:
        bond = 0.15

    return bond


class YoungLaplaceModel:
    _shape: Optional[YoungLaplaceShape] = None

    def __init__(self, data: Tuple[np.ndarray, np.ndarray]) -> None:
        self.data = np.copy(data)
        self.data.flags.writeable = False

        self._params = np.empty(len(YoungLaplaceParam))
        self._params_set = False
        self._s = np.empty(shape=(self.data.shape[1],))
        self._residuals = np.empty(shape=(self.data.shape[1],))
        self._jac = np.empty(shape=(self.data.shape[1], len(self._params)))

    def set_params(self, params: Sequence[float]) -> None:
        if self._params_set and (self._params == params).all():
            return

        bond = params[YoungLaplaceParam.BOND]
        radius = params[YoungLaplaceParam.RADIUS]
        X0 = params[YoungLaplaceParam.APEX_X]
        Y0 = params[YoungLaplaceParam.APEX_Y]
        w = params[YoungLaplaceParam.ROTATION]

        s = self._s

        residuals = self._residuals
        de_dBo = self._jac[:, YoungLaplaceParam.BOND]
        de_dR = self._jac[:, YoungLaplaceParam.RADIUS]
        de_dX0 = self._jac[:, YoungLaplaceParam.APEX_X]
        de_dY0 = self._jac[:, YoungLaplaceParam.APEX_Y]
        de_dw = self._jac[:, YoungLaplaceParam.ROTATION]

        shape = self._get_shape(bond)
        Q = rotation_mat2d(w)

        data_x, data_y = self.data
        data_r, data_z = Q.T @ (data_x - X0, data_y - Y0)

        s[:] = shape.closest(data_r / radius, data_z / radius)
        r, z = radius * shape(s)
        dr_dBo, dz_dBo = radius * shape.DBo(s)
        e_r = data_r - r
        e_z = data_z - z
        e = np.hypot(e_r, e_z)

        # Set residues for points inside the drop as negative and outside as positive.
        e[np.signbit(e_r) != np.signbit(r)] *= -1

        residuals[:] = e
        de_dBo[:] = -(e_r * dr_dBo + e_z * dz_dBo) / e  # derivative w.r.t. Bond number
        de_dR[:] = -(e_r * r + e_z * z) / (radius * e)  # derivative w.r.t. radius
        # derivative w.r.t. apex (x, y)-coordinates
        de_dX0[:], de_dY0[:] = -Q @ (e_r, e_z) / e
        # derivative w.r.t. rotation
        de_dw[:] = (e_r * z - e_z * r) / e

        self._params[:] = params

    def _get_shape(self, bond: float) -> YoungLaplaceShape:
        if self._shape is None or self._shape.bond != bond:
            self._shape = YoungLaplaceShape(bond)

        return self._shape

    @property
    def params(self) -> Sequence[int]:
        params = self._params[:]
        params.flags.writeable = False
        return params

    @property
    def dof(self) -> int:
        return self.data.shape[1] - len(self.params) + 1

    @property
    def jac(self) -> np.ndarray:
        jac = self._jac[:]
        jac.flags.writeable = False
        return jac

    @property
    def residuals(self) -> np.ndarray:
        residuals = self._residuals[:]
        residuals.flags.writeable = False
        return residuals

    @property
    def closest(self) -> np.ndarray:
        xy = np.empty_like(self.data, dtype=float)

        bond = self._params[YoungLaplaceParam.BOND]
        radius = self._params[YoungLaplaceParam.RADIUS]
        X0 = self._params[YoungLaplaceParam.APEX_X]
        Y0 = self._params[YoungLaplaceParam.APEX_Y]
        w = self._params[YoungLaplaceParam.ROTATION]

        shape = self._get_shape(bond)
        Q = rotation_mat2d(w)
        s = self._s

        rz = radius * shape(s)
        xy[:] = Q @ rz + [[X0], [Y0]]

        return xy

    @property
    def arclengths(self) -> np.ndarray:
        s = self._s[:]
        s.flags.writeable = False
        return s

    @property
    def volume(self) -> float:
        bond = self._params[YoungLaplaceParam.BOND]
        radius = self._params[YoungLaplaceParam.RADIUS]

        shape = self._get_shape(bond)
        s = self._s

        return radius**3 * shape.volume(s.max())

    @property
    def surface_area(self) -> float:
        bond = self._params[YoungLaplaceParam.BOND]
        radius = self._params[YoungLaplaceParam.RADIUS]

        shape = self._get_shape(bond)
        s = self._s

        return radius**2 * shape.surface_area(s.max())


class YoungLaplaceParam(IntEnum):
    BOND = 0
    RADIUS = auto()
    APEX_X = auto()
    APEX_Y = auto()
    ROTATION = auto()



from modules.core.classes import ExperimentalSetup
from modules.image.select_regions import (
    user_roi,
    set_scale,
    set_screen_position,
    user_select_region,
)
from modules.ift.younglaplace.younglaplace import young_laplace_fit
from modules.ift.younglaplace.shape import YoungLaplaceShape
from modules.ift.pendant import extract_pendant_features, analyze_ift
from utils.misc import rotation_mat2d
from utils.enums import RegionSelect
from utils.config import MAX_ARCLENGTH
from utils.geometry import Rect2

from PIL import Image
from typing import Callable
import cv2
import os
import csv
import numpy as np
import timeit

# from modules.PlotManager import PlotManager


class IftDataProcessor:
    def process_data(
        self, user_input_data: ExperimentalSetup, callback: Callable = None
    ):

        n_frames = user_input_data.number_of_frames
        time = 0

        for i in range(len(user_input_data.import_files)):
            # Load the image (assuming OpenCV)
            image = user_input_data.import_files[i]
            if image is None:
                print(f"Failed to load image: {image}")
                continue
            print("\nProcessing frame %d of %d..." % (i + 1, n_frames))
            input_file = user_input_data.import_files[i]
            print("\nProcessing " + input_file)
            time_start = timeit.default_timer()
            # 1. Extract drop and needle regions
            # drop_points, needle_diameter_px, drop_region, needle_region, image, drop_image, needle_fit_result = extract_pendant_features(image)
            print(user_input_data.fit_result[i])
            analyzed_ift = analyze_ift(
                user_input_data.fit_result[i],
                drop_density=user_input_data.drop_density,
                continuous_density=user_input_data.density_outer,
                needle_diameter_mm=user_input_data.needle_diameter_mm,
                needle_diameter_px=user_input_data.needle_diameter_px[i],
            )
            self.draw_fitted_shape(user_input_data, i, image)
            time_end = timeit.default_timer()
            duration = time_end - time_start
            analyzed_ift[5] = time + i * user_input_data.frame_interval
            # Save the analyzed IFT results
            # print("Analyzed IFT:", analyzed_ift)
            user_input_data.ift_results[i] = analyzed_ift

            print("Time taken for frame %d: %.2f seconds" % (i + 1, duration))
            print("callback: ", i)

        if callback:
            callback(user_input_data)

    def process_preparation(self, user_input_data: ExperimentalSetup):
        n_frames = user_input_data.number_of_frames
        # Initialize drop_images if not already
        print(
            "####################################: ", len(
                user_input_data.import_files)
        )
        user_input_data.drop_images = ["None"] * n_frames
        # num_of_images = len(user_input_data.import_files)
        user_input_data.drop_points = ["None"] * n_frames
        user_input_data.needle_diameter_px = ["None"] * n_frames
        user_input_data.drop_region = ["None"] * n_frames
        user_input_data.needle_region = ["None"] * n_frames
        user_input_data.fit_result = ["None"] * n_frames
        user_input_data.ift_results = ["None"] * n_frames
        user_input_data.drop_contour_images = ["None"] * n_frames
        user_input_data.processed_images = ["None"] * n_frames

        for i in range(len(user_input_data.import_files)):
            print("\nProcessing frame %d of %d..." % (i + 1, n_frames))
            input_file = user_input_data.import_files[i]
            print("\nProcessing " + input_file)
            time_start = timeit.default_timer()
            # Load the image (assuming OpenCV)
            image_file = user_input_data.import_files[i]
            drop_region = None
            needle_region = None

            if image_file is None:
                print(f"Failed to load image: {input_file}")
                continue

            image = cv2.imread(image_file, cv2.IMREAD_COLOR)
            if image is None:
                print(f"Could not load image at {image_file}")

            screen_size = user_input_data.screen_resolution
            image_size = image.shape
            scale = set_scale(image_size, screen_size)
            screen_position = set_screen_position(screen_size)

            if user_input_data.drop_id_method == RegionSelect.USER_SELECTED:
                print("Select drop region for Image {i}")
                [(min_x, min_y), (max_x, max_y)], _ = user_select_region(
                    image, f"Select drop region for Image {i}", scale, screen_position
                )
                drop_region = Rect2(int(min_x), int(min_y),
                                    int(max_x), int(max_y))
                print("Drop region: ", drop_region)

            if user_input_data.needle_region_method == RegionSelect.USER_SELECTED:
                print("Select needle region for Image {i}")
                [(min_x, min_y), (max_x, max_y)], _ = user_select_region(
                    image, f"Select needle region for Image {i}", scale, screen_position
                )
                needle_region = Rect2(int(min_x), int(
                    min_y), int(max_x), int(max_y))
                print("Needle region: ", needle_region)

            (
                drop_points,
                needle_diameter_px,
                drop_region,
                needle_region,
                image,
                drop_image,
                needle_fit_result,
            ) = extract_pendant_features(image, drop_region, needle_region)
            user_input_data.fit_result[i] = young_laplace_fit(
                drop_points, verbose=True)
            user_input_data.drop_points[i] = drop_points
            user_input_data.needle_diameter_px[i] = needle_diameter_px
            user_input_data.drop_region[i] = drop_region
            user_input_data.needle_region[i] = needle_region
            self.draw_regions(user_input_data, i, image)

    def draw_regions(
        self, user_input_data: ExperimentalSetup, i: int, image: np.ndarray
    ):
        drop_region = user_input_data.drop_region[i]
        needle_region = user_input_data.needle_region[i]
        regions_image = image.copy()
        # Draw drop_region (blue)
        if drop_region is not None:
            regions_image = cv2.rectangle(
                regions_image,
                (int(drop_region.x0), int(drop_region.y0)),
                (int(drop_region.x1), int(drop_region.y1)),
                (255, 0, 0),
                2,
            )
        # Draw needle_region (red)
        if needle_region is not None:
            regions_image = cv2.rectangle(
                regions_image,
                (int(needle_region.x0), int(needle_region.y0)),
                (int(needle_region.x1), int(needle_region.y1)),
                (0, 0, 255),
                2,
            )
        image_pil = Image.fromarray(
            cv2.cvtColor(regions_image, cv2.COLOR_BGR2RGB))
        user_input_data.processed_images[i] = image_pil

    def draw_fitted_shape(
        self, user_input_data: ExperimentalSetup, drop_index: int, image: np.ndarray
    ):
        fit_result = user_input_data.fit_result[drop_index]
        shape = YoungLaplaceShape(fit_result.bond)
        _ = shape(MAX_ARCLENGTH)
        # 2. Pick arclength values to sample.  We’ll reuse the ones from the fit:
        s_values = fit_result.arclengths
        # 3. Generate (r, z) coordinates pointwise
        #    shape(s) returns a length‐2 array [r, z]
        rz = np.array([shape(s) for s in s_values])  # shape (N,2)
        r_coords = rz[:, 0]
        z_coords = rz[:, 1]
        # 4. Now transform into image coords:
        #    a) scale by fitted radius
        rz_scaled = fit_result.radius * rz.T  # shape (2, N)
        #    b) rotate by fitted rotation
        Q = rotation_mat2d(fit_result.rotation)
        xy_fitted = Q @ rz_scaled  # still (2, N)
        #    c) translate to apex location
        apex = np.array([fit_result.apex_x, fit_result.apex_y]).reshape(2, 1)
        xy_fitted += apex
        # Crop the original BGR image using drop_region
        drop_region = user_input_data.drop_region[drop_index]
        y0 = int(drop_region.y0)
        y1 = int(drop_region.y1)
        x0 = int(drop_region.x0)
        x1 = int(drop_region.x1)
        image = cv2.imread(image)

        # Make a copy to draw on (still BGR)
        img_to_draw_on = image
        # Translate fitted points to be relative to the cropped image
        translated_fitted_x = xy_fitted[0, :]
        translated_fitted_y = xy_fitted[1, :]
        # Draw the fitted shape as individual points on the cropped BGR image
        point_radius = 0  # As in your existing code
        point_color_bgr = (0, 0, 255)
        point_thickness = -1
        for i in range(len(translated_fitted_x)):
            x_coord = int(translated_fitted_x[i])
            y_coord = int(translated_fitted_y[i])
            cv2.circle(
                img_to_draw_on,
                (x_coord, y_coord),
                point_radius,
                point_color_bgr,
                point_thickness,
            )

        # (cross-platform)
        save_dir = os.path.join(
            os.path.join(os.path.expanduser("~")),
            "OpenDrop",
            "outputs",
            "contour_images"
        )
        os.makedirs(save_dir, exist_ok=True)

        original_path = user_input_data.import_files[drop_index]
        original_name = os.path.basename(original_path)

        save_path = os.path.join(save_dir, original_name)
        cv2.imwrite(save_path, img_to_draw_on)

        user_input_data.drop_contour_images[drop_index] = save_path

    def save_result(self, user_input_data: ExperimentalSetup, output_file_path: str):
        """
        Save experiment results to a CSV file with columns:
        Filename, Time, IFT, V, SA, Bond, Worth
        """
        with open(output_file_path, "w", newline="") as csvfile:
            writer = csv.writer(csvfile)
            writer.writerow(
                [
                    "Filename",
                    "Time",
                    "IFT (mN/m)",
                    "Volume (mm^3)",
                    "Surface Area (mm^2)",
                    "Bond",
                    "Worth",
                ]
            )
            for i, result in enumerate(user_input_data.ift_results):
                if result is not None and result != "None":
                    # result: [IFT, V, SA, Bond, Worth, Time]
                    writer.writerow(
                        [
                            user_input_data.import_files[i],
                            f"{result[5]:.1f}",
                            f"{result[0]:.1f}",
                            f"{result[1]:.2f}",
                            f"{result[2]:.2f}",
                            f"{result[3]:.4f}",
                            f"{result[4]:.4f}",
                        ]
                    )



#!/usr/bin/env python
# coding=utf-8
from modules.core.classes import ExperimentalDrop
from utils.config import LEFT_ANGLE, RIGHT_ANGLE
from utils.enums import FittingMethod

# from __future__ import print_function


def perform_fits(
    experimental_drop: ExperimentalDrop,
    tangent=False,
    polynomial=False,
    circle=False,
    ellipse=False,
    yl=False,
):
    if tangent == True:
        from modules.fitting.polynomial_fit import polynomial_fit

        tangent_angles, tangent_CPs, tangent_lines, tangent_errors, tangent_timings = (
            polynomial_fit(experimental_drop.drop_contour, polynomial_degree=1)
        )
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT] = {}
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][LEFT_ANGLE] = (
            tangent_angles[0]
        )
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][RIGHT_ANGLE] = (
            tangent_angles[1]
        )
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][
            "contact points"
        ] = tangent_CPs
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][
            "tangent lines"
        ] = tangent_lines
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][
            "errors"
        ] = tangent_errors
        experimental_drop.contact_angles[FittingMethod.TANGENT_FIT][
            "timings"
        ] = tangent_timings

    if polynomial == True:
        from modules.fitting.polynomial_fit import polynomial_fit

        (
            polynomial_angles,
            polynomial_CPs,
            polynomial_lines,
            polynomial_errors,
            polynomial_timings,
        ) = polynomial_fit(experimental_drop.drop_contour, polynomial_degree=2)
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT] = {}
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][LEFT_ANGLE] = (
            polynomial_angles[0]
        )
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][RIGHT_ANGLE] = (
            polynomial_angles[1]
        )
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][
            "contact points"
        ] = polynomial_CPs
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][
            "tangent lines"
        ] = polynomial_lines
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][
            "errors"
        ] = polynomial_errors
        experimental_drop.contact_angles[FittingMethod.POLYNOMIAL_FIT][
            "timings"
        ] = polynomial_timings

    if circle == True:
        from modules.fitting.circular_fit import circular_fit

        (
            circle_angles,
            circle_center,
            circle_radius,
            circle_intercepts,
            circle_errors,
            circle_timings,
        ) = circular_fit(experimental_drop.drop_contour)
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT] = {}
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][LEFT_ANGLE] = (
            circle_angles[0]
        )
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][RIGHT_ANGLE] = (
            circle_angles[1]
        )
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][
            "baseline intercepts"
        ] = circle_intercepts
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][
            "circle center"
        ] = circle_center
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][
            "circle radius"
        ] = circle_radius
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][
            "errors"
        ] = circle_errors
        experimental_drop.contact_angles[FittingMethod.CIRCLE_FIT][
            "timings"
        ] = circle_timings

    if ellipse == True:
        from modules.fitting.ellipse_fit import ellipse_fit

        (
            ellipse_angles,
            ellipse_intercepts,
            ellipse_center,
            ellipse_ab,
            ellipse_rotation,
            ellipse_errors,
            ellipse_timings,
        ) = ellipse_fit(experimental_drop.drop_contour)
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT] = {}
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][LEFT_ANGLE] = (
            ellipse_angles[0]
        )
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][RIGHT_ANGLE] = (
            ellipse_angles[1]
        )
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "baseline intercepts"
        ] = ellipse_intercepts
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "ellipse center"
        ] = ellipse_center
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "ellipse a and b"
        ] = ellipse_ab
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "ellipse rotation"
        ] = ellipse_rotation
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "errors"
        ] = ellipse_errors
        experimental_drop.contact_angles[FittingMethod.ELLIPSE_FIT][
            "timings"
        ] = ellipse_timings

    if yl == True:
        from modules.fitting.BA_fit import yl_fit

        (
            yl_angles,
            yl_bo,
            yl_baselinewidth,
            yl_volume,
            yl_shape,
            yl_baseline,
            yl_errors,
            sym_errors,
            yl_timings,
        ) = yl_fit(experimental_drop.drop_contour)
        experimental_drop.contact_angles[FittingMethod.YL_FIT] = {}
        experimental_drop.contact_angles[FittingMethod.YL_FIT][LEFT_ANGLE] = yl_angles[
            0
        ]
        experimental_drop.contact_angles[FittingMethod.YL_FIT][RIGHT_ANGLE] = yl_angles[
            1
        ]
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["bond number"] = yl_bo
        experimental_drop.contact_angles[FittingMethod.YL_FIT][
            "baseline width"
        ] = yl_baselinewidth
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["volume"] = yl_volume
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["fit shape"] = yl_shape
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["baseline"] = yl_baseline
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["errors"] = yl_errors
        experimental_drop.contact_angles[FittingMethod.YL_FIT][
            "symmetry errors"
        ] = sym_errors
        experimental_drop.contact_angles[FittingMethod.YL_FIT]["timings"] = yl_timings



from modules.fitting.BA_fit import (
    cluster_optics,
    distance1,
    dist,
    optimized_path,
    prepare_hydrophobic,
    bashforth_adams,
    sim_bashforth_adams,
    fit_bashforth_adams,
    calculate_angle,
    fit_circle,
    generate_circle_vectors,
    find_intersection,
    yl_fit,
)
import pytest
import numpy as np
import warnings
from unittest.mock import patch, MagicMock
import sys
import os

# import cv2


sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

warnings.filterwarnings("ignore", category=DeprecationWarning)

# ===== Fixtures =====


@pytest.fixture
def sample_data():
    return np.array([[1, 1], [1.1, 1.1], [0, 5], [0.1, 5.1], [5, 0], [5.1, 0.1]])


@pytest.fixture
def sample_profile():
    return np.array([[0, 0], [1, 1], [2, 2], [3, 1], [4, 0]])


# ===== Tests =====


def test_cluster_optics(sample_data):
    result = cluster_optics(sample_data, xi=0.05)
    assert isinstance(result, dict)
    assert len(result) > 0


def test_distance1():
    assert distance1([0, 0], [3, 4]) == pytest.approx(5.0)


def test_dist():
    params = [0, 0, 1]
    points = np.array([[1, 0], [0, 1], [-1, 0], [0, -1]])
    assert dist(params, points) == pytest.approx(0.0)


def test_optimized_path(sample_data):
    path = optimized_path(sample_data)
    assert len(path) <= len(sample_data)
    for i in range(1, len(path)):
        assert distance1(path[i - 1], path[i]) < 5


@patch("modules.fitting.BA_fit.plt.show")
def test_prepare_hydrophobic(mock_show, sample_profile):
    profile, CPs = prepare_hydrophobic(sample_profile, display=True)
    assert isinstance(profile, np.ndarray)
    assert isinstance(CPs, dict)


def test_bashforth_adams():
    result = bashforth_adams(0, [1, 1], 1, 1)
    assert isinstance(result, tuple)
    assert len(result) == 2


@patch("modules.fitting.BA_fit.solve_ivp")
def test_sim_bashforth_adams(mock_solve_ivp):
    mock_solve_ivp.return_value = MagicMock(
        t=np.array([0, 1]), y=np.array([[1, 1], [2, 2]])
    )
    angles, pred, Bo = sim_bashforth_adams(1, 1, 1)
    assert isinstance(angles, np.ndarray)
    assert isinstance(pred, np.ndarray)
    assert isinstance(Bo, float)


@patch("modules.fitting.BA_fit.opt.minimize")
def test_fit_bashforth_adams(mock_minimize, sample_profile):
    mock_minimize.return_value = MagicMock(x=[1, 1])
    result = fit_bashforth_adams(sample_profile)
    assert isinstance(result, MagicMock)


def test_calculate_angle():
    angle = calculate_angle(np.array([1, 0]), np.array([0, 1]))
    assert angle == pytest.approx(90.0)


@patch("modules.fitting.BA_fit.opt.minimize")
def test_fit_circle(mock_minimize, sample_profile):
    mock_minimize.return_value = {"x": [0, 0, 1], "fun": 0.1}
    result = fit_circle(sample_profile)
    assert isinstance(result, dict)


def test_generate_circle_vectors():
    v1, v2 = generate_circle_vectors([1, 0])
    assert np.allclose(v1, [-1, 0])
    assert np.allclose(v2, [0, 1])


def test_find_intersection():
    baseline_coeffs = [0, 1]
    circ_params = [0, 0, 1]
    x_t, y_t = find_intersection(baseline_coeffs, circ_params)
    assert x_t == pytest.approx(1.0)
    assert y_t == pytest.approx(0.0)


# @patch('modules.fitting.BA_fit.cv2.findContours')
# def test_find_contours(mock_findContours):
#     mock_findContours.return_value = ([np.array([[[0, 0]], [[1, 1]]])], None)
#     image = np.ones((10, 10), dtype=np.uint8)
#     contours = find_contours(image)
#     assert isinstance(contours, list)
#     assert len(contours) == 1
#     assert np.array_equal(contours[0], np.array([[0, 0], [1, 1]]))


@patch("modules.fitting.BA_fit.plt.show")
@patch("modules.fitting.BA_fit.fit_circle", return_value={"x": [0, 0, 1], "fun": 0.1})
@patch("modules.fitting.BA_fit.find_intersection", return_value=(0.5, 0.5))
@patch("modules.fitting.BA_fit.generate_circle_vectors", return_value=([1, 0], [0, 1]))
@patch("modules.fitting.BA_fit.calculate_angle", return_value=90)
@patch("modules.fitting.BA_fit.fit_bashforth_adams", return_value=MagicMock(x=[1, 1]))
@patch(
    "modules.fitting.BA_fit.sim_bashforth_adams",
    return_value=(np.array([0, 90]), np.array([[0, 0], [1, 1]]), 1.0),
)
def test_yl_fit(
    mock_sim,
    mock_fit,
    mock_calc,
    mock_gen,
    mock_find,
    mock_circle,
    mock_show,
    sample_profile,
):
    result = yl_fit(sample_profile, display=True)
    assert isinstance(result, tuple)
    assert len(result) == 9


# Optional: run directly
if __name__ == "__main__":
    import sys

    sys.exit(pytest.main(["-v", __file__]))



from utils.geometry import Vector2

from typing import Sequence, NamedTuple, Optional
from enum import IntEnum, auto
import numpy as np
import scipy.optimize

__all__ = (
    "CircleFitResult",
    "circle_fit",
)

DELTA_TOL = 1.0e-8
GRADIENT_TOL = 1.0e-8
OBJECTIVE_TOL = 1.0e-8


class CircleFitResult(NamedTuple):
    center: Vector2[float]
    radius: float

    objective: float
    residuals: np.ndarray


def circle_fit(
    data: np.ndarray,
    *,
    loss: str = "linear",
    f_scale: float = 1.0,
    xc: Optional[float] = None,
    yc: Optional[float] = None,
    radius: Optional[float] = None,
    verbose: bool = False,
) -> Optional[CircleFitResult]:
    if data.shape[1] == 0:
        return None

    model = CircleModel(data)

    def fun(params: Sequence[float]) -> np.ndarray:
        model.set_params(params)
        residuals = model.residuals.copy()
        return residuals

    def jac(params: Sequence[float]) -> np.ndarray:
        model.set_params(params)
        jac = model.jac.copy()
        return jac

    initial_params = np.empty(len(CircleParam))

    if xc is None or yc is None:
        xc, yc = data.mean(axis=1)

    if radius is None:
        tx, ty = data[0] - xc, data[1] - yc
        radius = np.median(np.sqrt(tx**2 + ty**2))

    initial_params[CircleParam.CENTER_X] = xc
    initial_params[CircleParam.CENTER_Y] = yc
    initial_params[CircleParam.RADIUS] = radius
    model.set_params(initial_params)

    try:
        optimize_result = scipy.optimize.least_squares(
            fun,
            model.params,
            jac,
            method="lm" if loss == "linear" else "trf",
            loss=loss,
            f_scale=f_scale,
            x_scale="jac",
            ftol=OBJECTIVE_TOL,
            xtol=DELTA_TOL,
            gtol=GRADIENT_TOL,
            max_nfev=50,
            verbose=2 if verbose else 0,
        )
    except ValueError:
        return None

    # Update model parameters to final result.
    model.set_params(optimize_result.x)

    result = CircleFitResult(
        center=Vector2(
            model.params[CircleParam.CENTER_X], model.params[CircleParam.CENTER_Y]
        ),
        radius=model.params[CircleParam.RADIUS],
        objective=(model.residuals**2).sum() / model.dof,
        residuals=model.residuals,
    )

    return result


class CircleModel:
    def __init__(self, data: np.ndarray) -> None:
        if data.flags.writeable:
            data = data.copy()
            data.flags.writeable = False

        self.data = data

        self._params = np.empty(len(CircleParam))
        self._params_set = False
        self._residuals = np.empty(shape=(self.data.shape[1],))
        self._jac = np.empty(shape=(self.data.shape[1], len(self._params)))
        self._lmask = np.empty(shape=(self.data.shape[1],), dtype=bool)

    def set_params(self, params: Sequence[float]) -> None:
        if self._params_set and (self._params == params).all():
            return

        xc = params[CircleParam.CENTER_X]
        yc = params[CircleParam.CENTER_Y]
        R = params[CircleParam.RADIUS]

        e = self._residuals
        de_dxc = self._jac[:, CircleParam.CENTER_X]
        de_dyc = self._jac[:, CircleParam.CENTER_Y]
        de_dR = self._jac[:, CircleParam.RADIUS]

        x, y = self.data
        tx = x - xc
        ty = y - yc
        r = np.sqrt(tx**2 + ty**2)

        e[:] = r - R
        de_dxc[:] = -tx / r
        de_dyc[:] = -ty / r
        de_dR[:] = -1

        self._params[:] = params

    @property
    def params(self) -> Sequence[int]:
        params = self._params[:]
        params.flags.writeable = False
        return params

    @property
    def dof(self) -> int:
        return self.data.shape[1] - len(self.params) + 1

    @property
    def jac(self) -> np.ndarray:
        jac = self._jac[:]
        jac.flags.writeable = False
        return jac

    @property
    def residuals(self) -> np.ndarray:
        residuals = self._residuals[:]
        residuals.flags.writeable = False
        return residuals


class CircleParam(IntEnum):
    CENTER_X = 0
    CENTER_Y = auto()
    RADIUS = auto()



#!/usr/bin/env python
# coding=utf-8

"""This code serves as a discrete instance of the polynomial fit method of
contact angle analysis.

Polynomial fit code taken from the most recent version of conan -
conan-ML_cv1.1/modules/select_regions.py"""

# Polynomial fit from the most recent version of conan - conan-ML_v1.1/modules/select_regions.py

from utils.config import CV2_VERSION

from sklearn.cluster import OPTICS  # for clustering algorithm
from scipy.spatial import distance
from scipy.integrate import solve_ivp
import scipy.optimize as opt
import numba
import math
import numpy as np
import matplotlib.pyplot as plt
import cv2
import time


def cluster_optics(sample, out_style="coords", xi=None, eps=None, verbose=0):
    """Takes an array (or list) of the form [[x1,y1],[x2,y2],...,[xn,yn]].
    Clusters are outputted in the form of a dictionary.

    If out_style='coords' each dictionary entry is a group, and points are outputted in coordinate form.
    If out_xy='xy' there are two dictionary entries for each group, one labeled as nx and one as ny
    (where n is the label of the group)

    If xi (float between 0 and 1) is not None and eps is None, then the xi clustering method is used.
    The optics algorithm defines clusters based on the minimum steepness on the reachability plot.
    For example, an upwards point in the reachability plot is defined by the ratio from one point to
    its successor being at most 1-xi.

    If eps (float) is not None and xi is None, then the dbscan clustering method is used. Where eps is the
    maximum distance between two samples for one to be considered as in the neighborhood of the other.

    https://stackoverflow.com/questions/47974874/algorithm-for-grouping-points-in-given-distance
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.OPTICS.html
    """
    if eps is not None and xi is None:
        clustering = OPTICS(min_samples=2, cluster_method="dbscan", eps=eps).fit(
            sample
        )  # cluster_method changed to dbscan (so eps can be set)
    elif xi is not None and eps is None:
        # original had xi = 0.05, xi as 0.1 in function input
        clustering = OPTICS(min_samples=2, xi=xi).fit(sample)
    else:
        raise ValueError(
            "only one of eps and xi can be chosen but not neither nor both"
        )
    groups = list(set(clustering.labels_))

    if verbose == 2:
        print(clustering.labels_)
    elif verbose == 1:
        print(groups)

    dic = {}
    for n in groups:
        if n not in dic:
            dic[n] = []
        for i in range(len(sample)):
            if clustering.labels_[i] == n:
                dic[n].append(sample[i])

    # separate points and graph
    dic2 = {}
    for k in dic.keys():
        x = []
        y = []
        for i in range(len(dic[k])):
            x.append(dic[k][i][0])
        dic2[str(k) + "x"] = x
        for i in range(len(dic[k])):
            y.append(dic[k][i][1])
        dic2[str(k) + "y"] = y

    if out_style == "coords":
        return dic
    elif out_style == "xy":
        return dic2


def distance1(p1, p2):
    """This function computes the distance between 2 points defined by
    P1 = (x1,y1) and P2 = (x2,y2)"""
    return ((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2) ** 0.5


def optimized_path(coords, start=None):
    """This function finds the nearest point to a point
    coords should be a list in this format coords = [ [x1, y1], [x2, y2] , ...]
    https://stackoverflow.com/questions/45829155/sort-points-in-order-to-have-a-continuous-curve-using-python
    """
    if isinstance(coords, list) == False:
        coords = coords.tolist()
    if 0:
        if isinstance(start, list) == False:
            try:
                start = start.tolist()
            except:
                start = list(start)
    if start is None:
        start = coords[0]
    pass_by = coords
    path = [start]
    pass_by.remove(start)
    while pass_by:
        nearest = min(pass_by, key=lambda x: distance1(path[-1], x))
        path.append(nearest)
        pass_by.remove(nearest)
    path = np.array(path)

    # if there are any large jumps in distance, there is likely a mistake
    # therefore, the points after this jump should be ignored
    if 1:
        dists = []
        for i, point in enumerate(path):
            if i < len(path) - 1:
                dists.append(distance1(path[i], path[i + 1]))
        jump_idx = []
        for i, dist in enumerate(dists):
            if dist > 5:
                jump_idx.append(i)
        if len(jump_idx) > 0:
            path = path[: jump_idx[0]]

    return path


def prepare_hydrophobic(coords, xi=0.8, cluster=False, display=False):
    """takes an array (n,2) of coordinate points, and returns the left and right halfdrops of the contour.
    xi determines the minimum steepness on the reachability plot that constitutes a cluster boundary of the
    clustering algorithm.
    deg is the degree of the polynomial used to describe the shape of the droplet.

    This code is adapted from the prepare module, but this version differs in that it assumes that the drop
    is hydrophobic."""
    coords = coords.astype(float)
    # scan for clusers to remove noise and circle from lensing effect
    ################  MAY NEED TO OPTIMIZE eps/xi TO FIND APPROPRIATE GROUPINGS  ####################
    if cluster:  # turn this off bc using synthetic drops without lensing effect
        input_contour = coords
        dic, dic2 = cluster_optics(input_contour, xi=xi), cluster_optics(
            input_contour, out_style="xy", xi=xi
        )

        # print("number of groups: ",len(list(dic.keys())))

        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(list(dic.keys())))))
        for k in dic.keys():
            plt.plot(dic2[str(k) + "x"], dic2[str(k) + "y"], "o", color=next(colors))
        plt.title(str(len(dic.keys())) + " groups found by clustering")
        plt.show()
        plt.close()
        maxkey = max(dic, key=lambda k: len(dic[k]))

        # print('key to longest dictionary entry is: ',maxkey)

        # take the longest group
        longest = dic[maxkey]

    # flip contour so that min and max values are correct
    for coord in coords:
        coord[1] = -coord[1]

    longest = coords

    # print("first few coordinates of the longest contour: ",longest[:3])

    xlongest = []
    ylongest = []
    for i in range(len(longest)):
        xlongest.append(longest[i][0])
        ylongest.append(longest[i][1])

    # print("first few x coordinates of the longest contour: ",xlongest[:3])
    # print("first few y coordinates of the longest contour: ",ylongest[:3])

    # Find a appropriate epsilon value for cluster_optics, this will remove noise in the bottom 10% of the drop
    # .   most importantly noise is reduced at contact points.

    # variables in this process are how much and what part of the top of the droplet we use to be representative of
    # the full contour, and whether we use the max(distance) between points or the average between points, or
    # a scalar value of either.

    xtop = []  # isolate top 90% of drop
    ytop = []
    percent = 0.3
    # print('Isolate the top ',100-(percent*100),'% of the contour:')
    for n, y in enumerate(ylongest):
        if y > min(ylongest) + (max(ylongest) - min(ylongest)) * percent:
            xtop.append(xlongest[n])
            ytop.append(y)
    xtop = np.array(xtop)
    ytop = np.array(ytop)

    top = []
    for n, x in enumerate(xtop):
        top.append([xtop[n], ytop[n]])
    top = np.array(top)
    top_array = optimized_path(top)

    dists = []  # find the average distance between consecutive points
    for n, co in enumerate(top_array):
        if 1 < n:
            a = top_array[n]
            dist = np.linalg.norm(top_array[n] - top_array[n - 1])
            dists.append(dist)

    if display:
        # print(dists)
        print()
        print("Max dist between points is: ", max(dists))
        print("Average dist between points is: ", sum(dists) / len(dists))
        print("20% over the Max dist is: ", max(dists) * 1.2)
        print()
        print("Sort using cluster_optics with an epsilon value of ", max(dists) * 1.2)

    # how epsilon is chosen here is important
    # eps = (sum(dists)/len(dists))*2 # eps is 2 times the average distance between points
    # eps is 2.5 times the average distance between points
    eps = (sum(dists) / len(dists)) * 1.5
    input_contour = longest
    dic, dic2 = cluster_optics(input_contour, eps=eps), cluster_optics(
        input_contour, out_style="xy", eps=eps
    )

    # print("number of groups: ",len(list(dic.keys())))
    if display:
        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(list(dic.keys())))))
        for k in dic.keys():
            plt.plot(dic2[str(k) + "x"], dic2[str(k) + "y"], "o", color=next(colors))
        plt.title("Groups found by clustering with epsilon value of " + str(eps))
        plt.show()
        plt.close()
    maxkey = max(dic, key=lambda k: len(dic[k]))
    longest = dic[maxkey]
    # print("first few coordinates of the longest contour: ",longest[:3])

    xlongest = []
    ylongest = []
    for i in range(len(longest)):
        xlongest.append(longest[i][0])
        ylongest.append(longest[i][1])

    outline = np.empty((len(xlongest), 2))
    for i in range(len(xlongest)):
        outline[i, [0]] = xlongest[i]
        outline[i, [1]] = ylongest[i]

    ############################

    # find the apex of the drop and split the contour into left and right sides

    xtop = []  # isolate top 90% of drop
    ytop = []
    # percent = 0.1 # already defined
    # print('isolate the top ',100-(percent*100),'% of the contour:')
    for n, y in enumerate(ylongest):
        if y > min(ylongest) + (max(ylongest) - min(ylongest)) * percent:
            xtop.append(xlongest[n])
            ytop.append(y)
    xapex = (max(xtop) + min(xtop)) / 2

    l_drop = []
    r_drop = []
    for n in longest:
        if n[0] <= xapex:
            l_drop.append(n)
        if n[0] >= xapex:
            r_drop.append(n)
    l_drop = np.array(l_drop)
    r_drop = np.array(r_drop)

    # transpose both half drops so that they both face right and the apex of both is at 0,0
    r_drop[:, 0] = r_drop[:, 0] - xapex
    l_drop[:, 0] = -l_drop[:, 0] + xapex

    if display:
        plt.plot(r_drop[:, [0]], r_drop[:, [1]], "b,")
        # plt.show()
        # plt.close()
        plt.plot(l_drop[:, [0]], l_drop[:, [1]], "r,")
        # plt.gca().set_aspect('equal', adjustable='box')
        # plt.xlim([470,530])
        # plt.ylim([-188,-190])
        plt.show()
        plt.close()

    #############################

    # the drop has been split in half

    # this system has a user input which gives a rough indication of the contact point and the surface line

    # isolate the bottom 5% of the contour near the contact point

    drops = {}
    counter = 0
    crop_drop = {}
    CPs = {}
    for halfdrop in [l_drop, r_drop]:
        new_halfdrop = sorted(
            halfdrop.tolist(), key=lambda x: (x[0], -x[1])
        )  # top left to bottom right
        new_halfdrop = optimized_path(new_halfdrop)  # [::-1]

        xnew_halfdrop = new_halfdrop[:, [0]].reshape(len(new_halfdrop[:, [0]]))
        ynew_halfdrop = new_halfdrop[:, [1]].reshape(len(new_halfdrop[:, [1]]))

        # isolate the bottom of the drop to help identify contact points (may not need to do this for all scenarios)
        bottom = []
        top = []  # will need this later
        # print('isolate the bottom ',percent*100,'% of the contour:') # percent defined above
        div_line_value = (
            min(new_halfdrop[:, [1]])
            + (max(new_halfdrop[:, [1]]) - min(new_halfdrop[:, [1]])) * percent
        )
        for n in new_halfdrop:
            if n[1] < div_line_value:
                bottom.append(n)
            else:
                top.append(n)

        bottom = np.array(bottom)
        top = np.array(top)

        xbottom = bottom[:, [0]].reshape(len(bottom[:, [0]]))
        ybottom = bottom[:, [1]].reshape(len(bottom[:, [1]]))
        xtop = top[:, [0]].reshape(len(top[:, [0]]))
        ytop = top[:, [1]].reshape(len(top[:, [1]]))

        # print('max x value of halfdrop is: ',max(xhalfdrop))

        if 1:  # plot the bottom 10% of the contour
            plt.plot(xbottom, ybottom, "b,")
            plt.title("bottom 10% of the contour")
            # plt.xlim([130,200])
            plt.show()
            plt.close()

        #### Continue here assuming that the drop is hydrophobic ####
        if 1:
            # order all halfdrop points using optimized_path (very quick but occasionally makes mistakes)

            xCP = min(xbottom)
            # yCP = min([coord[1] for coord in new_halfdrop if coord[0]==xCP])
            yCP = max([coord[1] for coord in bottom if coord[0] == xCP])
            CPs[counter] = [xCP, yCP]

            if display:  # check
                plt.plot(new_halfdrop[:, 0], new_halfdrop[:, 1])
                plt.show()
                plt.close()

            # remove surface line past the contact point
            index = new_halfdrop.tolist().index(CPs[counter])  # ?

            new_halfdrop = new_halfdrop[: index + 1]

            if 0:
                xCP_index = [i for i, j in enumerate(xnew_halfdrop) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(ynew_halfdrop) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                new_halfdrop = np.zeros((len(xnew_halfdrop), 2))
                for n in range(len(xnew_halfdrop)):
                    new_halfdrop[n, [0]] = xnew_halfdrop[n]
                    new_halfdrop[n, [1]] = ynew_halfdrop[n]
                # print('first 3 points of new_halfdrop are: ',new_halfdrop[:3])
                # print('length of new_halfdrop is: ',len(new_halfdrop))

                if xCP_index == yCP_index:
                    if (
                        new_halfdrop[xCP_index[0] + 1][1]
                        > new_halfdrop[xCP_index[0] - 1][1]
                    ):
                        new_halfdrop = new_halfdrop[xCP_index[0] :]
                    else:
                        new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                else:
                    raise_error = True
                    for x in xCP_index:
                        for y in yCP_index:
                            if x == y:
                                raise_error = False
                                xCP_index = [x]
                                yCP_index = [y]
                                # print('indexes of the CP are: ',xCP_index[0],', ',yCP_index[0])
                                if (
                                    new_halfdrop[xCP_index[0] + 1][1]
                                    > new_halfdrop[xCP_index[0] - 1][1]
                                ):
                                    new_halfdrop = new_halfdrop[xCP_index[0] :]
                                else:
                                    new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                    if raise_error == True:
                        print(
                            "The index of the contact point x value is: ",
                            new_halfdrop[xCP_index],
                        )
                        print(
                            "The index of the contact point y value is: ",
                            new_halfdrop[yCP_index],
                        )
                        raise "indexes of x and y values of the contact point are not the same"

        if 0:
            # order all halfdrop points using two-opt (the slower method)

            # before any ordering is done, chop off the surface line that is past the drop edge
            del_indexes = []
            for index, coord in enumerate(bottom):
                if coord[0] > max(xtop):
                    del_indexes.append(index)
                if coord[1] < yCP:
                    del_indexes.append(index)
            # halfdrop = np.delete(halfdrop,del_indexes)
            xbot = np.delete(bottom[:, 0], del_indexes)
            ybot = np.delete(bottom[:, 1], del_indexes)

            # bottom = np.empty((len(xbot),2))
            # for n, coord in enumerate(bottom):
            #    bottom[n,0] = xbot[n]
            #    bottom[n,1] = ybot[n]

            bottom = np.array(list(zip(xbot, ybot)))

            # print('shape of another_halfdrop is: '+ str(type(another_halfdrop)))
            print("first few points of halfdrop are: ", halfdrop[:3])
            if display:
                colors = iter(jet(np.linspace(0, 1, len(bottom))))
                for n, coord in enumerate(bottom):
                    plt.plot(bottom[n, 0], bottom[n, 1], "o", color=next(colors))
                plt.title("Cropped bottom of hafldrop")
                plt.show()
                plt.close()

            # order points using traveling salesman two_opt code
            bottom = bottom[::-1]  # start at the top
            print(
                "Starting first coordinate of bottom slice of halfdrop is: ", bottom[0]
            )
            # increase improvement_threshold from 0.1 to 0.01
            new_bot, _ = two_opt(bottom, 0.01)
            if new_bot[0, 1] < new_bot[-1, 1]:
                new_bot = new_bot[::-1]
            xbot, ybot = new_bot[:, [0]], new_bot[:, [1]]

            if display:  # display
                colors = iter(jet(np.linspace(0, 1, len(bottom))))
                for n, coord in enumerate(bottom):
                    plt.plot(new_bot[n, 0], new_bot[n, 1], "o", color=next(colors))
                plt.title("Cropped ordered new_bot of halfdrop (starting from blue)")
                plt.show()
                plt.close()

                plt.plot(new_bot[:, 0], new_bot[:, 1])
                plt.title("Cropped ordered new_bot of halfdrop (line)")
                plt.show()
                plt.close()

            # order the top 90% so that the y value decreases

            print(
                "Sorting top ", 100 - (percent * 100), "% of the contour by y value..."
            )
            new_top = sorted(list(top), key=lambda x: x[1], reverse=True)
            new_top = np.array(new_top)

            # remove surface by deleting points after the contact point
            xCP_indexs = [i for i, j in enumerate(xbot) if j == xCP]
            # print('xCP_index is: ',xCP_index)
            yCP_indexs = [i for i, j in enumerate(ybot) if j == yCP]
            # print('yCP_index is: ',yCP_index)

            for xCP_index in xCP_indexs:
                for yCP_index in yCP_indexs:
                    if xCP_index == yCP_index:
                        try:
                            if ybot[yCP_index + 2] > ybot[yCP_index - 1]:
                                new_bot = np.zeros((len(xbot[yCP_index:]), 2))
                                for n in range(len(new_bot)):
                                    new_bot[n, [0]] = xbot[xCP_index + n]
                                    new_bot[n, [1]] = ybot[yCP_index + n]
                            else:
                                new_bot = np.zeros((len(xbot[:yCP_index]), 2))
                                for n in range(len(new_bot)):
                                    new_bot[n, [0]] = xbot[n]
                                    new_bot[n, [1]] = ybot[n]
                        except:
                            try:
                                if ybot[yCP_index] > ybot[yCP_index - 2]:
                                    new_bot = np.zeros((len(xbot[yCP_index:]), 2))
                                    for n in range(len(new_bot)):
                                        new_bot[n, [0]] = xbot[xCP_index + n]
                                        new_bot[n, [1]] = ybot[yCP_index + n]
                                else:
                                    new_bot = np.zeros((len(xbot[:yCP_index]), 2))
                                    for n in range(len(new_bot)):
                                        new_bot[n, [0]] = xbot[n]
                                        new_bot[n, [1]] = ybot[n]
                            except:
                                print("xCP_indexs are: ", xCP_indexs)
                                print("yCP_indexs are: ", yCP_indexs)
                                raise "indexes of x and y values of the contact point are not the same"
            new_halfdrop = np.concatenate((new_top, new_bot))

        if 0:  # order the points so that the baseline can be removed
            # before any ordering is done, chop off the surface line that is past the drop edge
            del_indexes = []
            for index, coord in enumerate(halfdrop):
                if coord[0] > max(xtop):
                    del_indexes.append(index)
            # halfdrop = np.delete(halfdrop,del_indexes)
            xhalfdrop = np.delete(xhalfdrop, del_indexes)
            yhalfdrop = np.delete(yhalfdrop, del_indexes)
            # print('shape of another_halfdrop is: '+ str(type(another_halfdrop)))
            # print('first few points of halfdrop are: ',halfdrop[:3])

            # order half contour points
            xx, yy = sort_to_line(xhalfdrop, yhalfdrop)
            add_top = False
            # print('length of halfdrop is: ', len(halfdrop))
            # print('length of xbottom is: ', len(xbottom))

            # if xx[0]<1: # then graph starts at the top
            surface_past_drop_index = []
            for n, x in enumerate(xx):
                if x > max(xtop):
                    surface_past_drop_index.append(n)
                    # xx = xx[:max(xtop)point]
            # print('Indexes of contour points past drop: ',surface_past_drop_index)

            # if the sort method will not work
            # assumes that the error is on the surface somewhere, so uses bottom of contour
            if len(xx) < len(xhalfdrop):
                add_top = True
                print()
                print(
                    "sort_to_line is not utilising the full contour, alternate ordering method being used"
                )
                print("check bottom 10% of contour...")
                # this method is much slower than the above, so use as few points as possible
                bot_list = []
                for n in range(len(xbottom)):
                    if xbottom[n] < max(xtop):
                        bot_list.append([xbottom[n], ybottom[n]])
                bot_array = np.array(bot_list)
                # 119.8 seconds for 247 points
                new_order, _ = two_opt(bot_array)

                xbot, ybot = new_order[:, [0]], new_order[:, [1]]

                if display:
                    plt.plot(xbot, ybot, "b-")
                    plt.title("Bottom of half drop, new order")
                    plt.show()
                    plt.close()

                xCP_index = [i for i, j in enumerate(xbot) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(ybot) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                if 0:
                    new_bot = np.zeros((len(xbot), 2))
                    for n in range(len(xx)):
                        new_bot[n, [0]] = xbot[n]
                        new_bot[n, [1]] = ybot[n]
                # xbot[xCP_index[0]:]
                if xCP_index == yCP_index:
                    if ybot[yCP_index[0] + 1] > ybot[yCP_index[0] - 1]:
                        new_bot = np.zeros((len(xbot[yCP_index[0] :]), 2))
                        for n in range(len(new_bot)):
                            new_bot[n, [0]] = xbot[xCP_index[0] + n]
                            new_bot[n, [1]] = ybot[yCP_index[0] + n]
                    else:
                        new_bot = np.zeros((len(xbot[: yCP_index[0]]), 2))
                        for n in range(len(new_bot)):
                            new_bot[n, [0]] = xbot[n]
                            new_bot[n, [1]] = ybot[n]
                else:
                    raise "indexes of x and y values of the contact point are not the same"

                # combine new_bot with top_array to give the isolated drop contour without surface
                if 0:
                    top_array = np.zeros((len(xtop), 2))
                    for n in range(len(xtop)):
                        top_array[n, [0]] = xtop[n]
                        top_array[n, [1]] = ytop[n]

                new_halfdrop = np.concatenate((top, new_bot))

                # re-order to check that the error was at the surface line
                xx, yy = sort_to_line(new_halfdrop[:, [0]], new_halfdrop[:, [1]])
                if len(xx) < len(
                    new_halfdrop
                ):  # then the error was in the top 90% of the drop
                    print("Checking top 90% of contour...")
                    new_top, _ = two_opt(top)
                    new_halfdrop = np.concatenate((new_top, new_bot))
                    xx, yy = sort_to_line(new_halfdrop[:, [0]], new_halfdrop[:, [1]])

            else:  # if sort_to_line worked as expected
                # find the indexs of the contact point and chop off the ends
                xCP_index = [i for i, j in enumerate(xx) if j == xCP]
                # print('xCP_index is: ',xCP_index)
                yCP_index = [i for i, j in enumerate(yy) if j == yCP]
                # print('yCP_index is: ',yCP_index)

                new_halfdrop = np.zeros((len(xx), 2))
                for n in range(len(xx)):
                    new_halfdrop[n, [0]] = xx[n]
                    new_halfdrop[n, [1]] = yy[n]
                # print('first 3 points of new_halfdrop are: ',new_halfdrop[:3])
                # print('length of new_halfdrop is: ',len(new_halfdrop))

                if xCP_index == yCP_index:
                    if (
                        new_halfdrop[xCP_index[0] + 1][1]
                        > new_halfdrop[xCP_index[0] - 1][1]
                    ):
                        new_halfdrop = new_halfdrop[xCP_index[0] :]
                    else:
                        new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                else:
                    raise_error = True
                    for x in xCP_index:
                        for y in yCP_index:
                            if x == y:
                                raise_error = False
                                xCP_index = [x]
                                yCP_index = [y]
                                # print('indexes of the CP are: ',xCP_index[0],', ',yCP_index[0])
                                if (
                                    new_halfdrop[xCP_index[0] + 1][1]
                                    > new_halfdrop[xCP_index[0] - 1][1]
                                ):
                                    new_halfdrop = new_halfdrop[xCP_index[0] :]
                                else:
                                    new_halfdrop = new_halfdrop[: xCP_index[0] + 1]
                    if raise_error == True:
                        print(
                            "The index of the contact point x value is: ",
                            new_halfdrop[xCP_index],
                        )
                        print(
                            "The index of the contact point y value is: ",
                            new_halfdrop[yCP_index],
                        )
                        raise "indexes of x and y values of the contact point are not the same"

        if counter == 0:
            drops[counter] = new_halfdrop[::-1]
        else:
            drops[counter] = new_halfdrop

        if display:  # display
            jet = plt.get_cmap("jet")
            colors = iter(jet(np.linspace(0, 1, len(new_halfdrop))))
            for k in new_halfdrop:
                plt.plot(k[0], k[1], "o", color=next(colors))
            plt.title("outputted halfdrop")
            plt.axis("equal")
            plt.show()
            plt.close()

        counter += 1

    # reflect the left drop and combine left and right

    profile = np.empty((len(drops[0]) + len(drops[1]), 2))
    for i, n in enumerate(drops[0]):
        flipped = n
        flipped[0] = -flipped[0]
        profile[i] = flipped
    for i, n in enumerate(drops[1]):
        profile[len(drops[0]) + i] = n
    CPs[0][0] = -CPs[0][0]

    if display:
        jet = plt.get_cmap("jet")
        colors = iter(jet(np.linspace(0, 1, len(profile))))
        for k in profile:
            plt.plot(k[0], k[1], "o", color=next(colors))
        plt.title("final output")
        # plt.plot(profile[:,0],profile[:,1],'b')
        plt.show()
        plt.close()

        plt.title("final output")
        plt.plot(profile[:, 0], profile[:, 1], "b")
        plt.show()
        plt.close()

    # flip upside down again so that contour follows image indexing
    # and transform to the right so that x=0 is no longer in line with apex
    for coord in profile:
        coord[1] = -coord[1]
        coord[0] = coord[0] + xapex
    for n in [0, 1]:
        CPs[n][1] = -CPs[n][1]
        CPs[n][0] = CPs[n][0] + xapex

    # flip original contour back to original orientation
    for coord in coords:
        coord[1] = -coord[1]

    return profile, CPs


def find_contours(image):
    """
    Calls cv2.findContours() on passed image in a way that is compatible with OpenCV 4.x, 3.x or 2.x
    versions. Passed image is a numpy.array.

    Note, cv2.findContours() will treat non-zero pixels as 1 and zero pixels as 0, so the edges detected will only
    be those on the boundary of pixels with non-zero and zero values.

    Returns a numpy array of the contours in descending arc length order.
    """
    if len(image.shape) > 2:
        raise ValueError(f"image {image} must be a single channel image")

    if CV2_VERSION >= (4, 0, 0):
        # In OpenCV 4.0, cv2.findContours() no longer returns three arguments, it reverts to the same return signature
        # as pre 3.2.0.
        contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )
    elif CV2_VERSION >= (3, 2, 0):
        # In OpenCV 3.2, cv2.findContours() does not modify the passed image and instead returns the
        # modified image as the first, of the three, return values.
        _, contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )
    else:
        contours, hierarchy = cv2.findContours(
            image, cv2.RETR_LIST, cv2.CHAIN_APPROX_TC89_KCOS
        )

    # Each contour has shape (n, 1, 2) where 'n' is the number of points. Presumably this is so each
    # point is a size 2 column vector, we don't want this so reshape it to a (n, 2)
    contours = [contour.reshape(contour.shape[0], 2) for contour in contours]

    # Sort the contours by arc length, descending order
    contours.sort(key=lambda c: cv2.arcLength(c, False), reverse=True)

    return contours


def extract_edges_CV(img):
    """
    give the image and return a list of [x.y] coordinates for the detected edges

    """
    IGNORE_EDGE_MARGIN = 1
    img = img.astype("uint8")
    try:
        gray = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    except:
        gray = img
    # ret, thresh = cv2.threshold(gray,threshValue,255,cv2.THRESH_BINARY)
    ret, thresh = cv2.threshold(gray, 0, 255, cv2.THRESH_BINARY + cv2.THRESH_OTSU)
    contours, hierarchy = cv2.findContours(thresh, cv2.RETR_LIST, cv2.CHAIN_APPROX_NONE)
    # Each contour has shape (n, 1, 2) where 'n' is the number of points. Presumably this is so each
    # point is a size 2 column vector, we don't want this so reshape it to a (n, 2)
    contours = [contour.reshape(contour.shape[0], 2) for contour in contours]

    # Sort the contours by arc length, descending order
    contours.sort(key=lambda c: cv2.arcLength(c, False), reverse=True)

    # Assume that the drop is the largest contour
    # drop_profile = contours[0]
    drop_profile = contours[0]

    # Put the drop contour coordinates in order (from ?? to ??)
    # drop_profile = squish_contour(drop_profile)

    # Ignore points of the drop profile near the edges of the drop image
    width, height = img.shape[1::-1]
    if not (width < IGNORE_EDGE_MARGIN or height < IGNORE_EDGE_MARGIN):
        mask = (
            (IGNORE_EDGE_MARGIN < drop_profile[:, 0])
            & (drop_profile[:, 0] < width - IGNORE_EDGE_MARGIN)
            & (IGNORE_EDGE_MARGIN < drop_profile[:, 1])
            & (drop_profile[:, 1] < height - IGNORE_EDGE_MARGIN)
        )
        drop_profile = drop_profile[mask]

    output = []
    for coord in drop_profile:
        if list(coord) not in output:
            output.append(list(coord))
    output = np.array(output)
    return output


def polynomial_closest_point(xp, yp, poly_points, display=False):
    """
    xp (float): The x-coordinate of the reference point
    yp (float): The y-coordinate of the reference point
    poly_points (array): The array of x, y coordinates outputted by the polynomial fit
    display (Boolean): Set to True to output figures and information.

    Returns:
        The distance between the reference point and the polynomial fit, and
        the coordinates of the closest point on the polynomial fit.

    """

    x = poly_points[:, 0]
    y = poly_points[:, 1]

    dist = np.sqrt((x - xp) ** 2 + (y - yp) ** 2)
    idx = list(dist).index(min(dist))

    # ddistdt = ((b ** 2 - a ** 2) * np.cos(t) + a * np.sin(np.deg2rad(th)) * yp - a * np.sin(np.deg2rad(th)) * yc + a * np.cos(np.deg2rad(th)) * xp - a * np.cos(np.deg2rad(th)) * xc) * np.sin(t) + ((-b * np.cos(np.deg2rad(th)) * yp) + b * np.cos(np.deg2rad(th)) * yc + b * np.sin(np.deg2rad(th)) * xp - b * np.sin(np.deg2rad(th)) * xc) * np.cos(t)
    # idx = np.where(ddistdt[1:] * ddistdt[:-1] < 0)[0]  # find zeros
    # m = (ddistdt[idx + 1] - ddistdt[idx]) / (t[idx + 1] - t[idx])  # slope
    if display:
        plt.figure(1)
        plt.plot(x, y, "-", xp, yp, "r+", x[idx], y[idx], "r+")
        plt.xlabel("x")
        plt.ylabel("y")
        plt.title("Circle, Point, and Zeros")

        plt.figure(2)
        for t, d in enumerate(dist):
            plt.plot(t, d, "m.")
        plt.plot(idx, dist[idx], "cx")
        plt.xlabel("index value of list")
        plt.ylabel("Distance")
        plt.title("Distance Function")

        print(f"xp: {xp}, x[idx]: {x[idx]}")
        print(f"yp: {yp}, y[idx]: {y[idx]}")
        print("Error is: ", dist[idx])

        plt.show()
        plt.close()

    return dist[idx], [x[idx], y[idx]]


def polynomial_fit_errors(pts1, pts2, fit_left, fit_right, display=False):
    """
    Calculates the minimum distance between a point and a point of the polynomial fit.

    Parameters:
        pts1 (array): The array of x, y coordindate points for the points near the left contact point
        pts2 (array): The array of x, y coordindate points for the points near the right contact point
        fit_left (np.polyfit): The fit found by np.polyfit for pts1
        fit_right (np.polyfit): The fit found by np.polyfit for pts2
        display (boolean): Set to true to show figures.

    Returns:
        dictionary: The MAE, MSE, RMSE, and maximum error of the contour as compared against the
        polynomial fit.
    """
    highresx_left = np.linspace(pts1[0, 0], pts1[-1, 0], 5 * pts1.shape[0])
    highresx_right = np.linspace(pts2[0, 0], pts2[-1, 0], 5 * pts2.shape[0])
    highres_fit_left = fit_left(highresx_left)
    highres_fit_right = fit_right(highresx_right)

    poly_points_left = np.array(list(zip(highresx_left, highres_fit_left)))
    poly_points_right = np.array(list(zip(highresx_right, highres_fit_right)))

    error_measures = {}

    # for left
    errors_left = []
    for point in pts1:
        dist2edge, edge_point = polynomial_closest_point(
            point[0], point[1], poly_points_left, display=display
        )
        errors_left.append(dist2edge)

    error_measures["MAE left"] = sum([abs(error) for error in errors_left]) / len(
        errors_left
    )
    error_measures["MSE left"] = sum([error**2 for error in errors_left]) / len(
        errors_left
    )
    error_measures["RMSE left"] = np.sqrt(
        sum([error**2 for error in errors_left]) / len(errors_left)
    )
    error_measures["Maximum error left"] = max(errors_left)

    # for right
    errors_right = []
    for point in pts2:
        dist2edge, edge_point = polynomial_closest_point(
            point[0], point[1], poly_points_right, display=display
        )
        errors_right.append(dist2edge)

    error_measures["MAE right"] = sum([abs(error) for error in errors_right]) / len(
        errors_right
    )
    error_measures["MSE right"] = sum([error**2 for error in errors_right]) / len(
        errors_right
    )
    error_measures["RMSE right"] = np.sqrt(
        sum([error**2 for error in errors_right]) / len(errors_right)
    )
    error_measures["Maximum error right"] = max(errors_right)

    error_measures["MAE"] = (
        sum([error_measures["MAE left"], error_measures["MAE right"]]) / 2
    )
    error_measures["MSE"] = (
        sum([error_measures["MSE left"], error_measures["MSE right"]]) / 2
    )
    error_measures["RMSE"] = (
        sum([error_measures["RMSE left"], error_measures["RMSE right"]]) / 2
    )
    error_measures["Maximum error"] = max(
        [error_measures["Maximum error left"], error_measures["Maximum error right"]]
    )

    return error_measures


def polynomial_fit_img(img, num_points=15, polynomial_degree=2, display=False):
    """Takes in input contact angle experimental image, and outputs the fitted
    angles and intercept coordinates. For best results, preprocessing should be
    performed before calling this function.

    The first and last num_points many points are are fitted to a polynomial, and contact
    angles are calculated by taking the gradient of the polynomial at the contact
    points. The tilt of the surface is accounted for by taking the gradient between
    contact points and balancing the polynomial angle against the surface gradient.

    Coords is the coordinates of the drop outline in the form of an array or list.
    num_points is the number of points included in the polynomial fit. The default number
        of points used is 10.
    display can be set to "True" to output figures."""

    # begin with method specific preprocessing of img data
    start_time = time.time()

    # array of x,y coords where lines are detected
    edges_pts = extract_edges_CV(img)

    if display:
        plt.imshow(img)
        plt.plot(edges_pts[:, 0], edges_pts[:, 1])
        plt.title("drop found by hough transform")
        plt.show()
        plt.close()

    tangent_drop, CPs = prepare_hydrophobic(edges_pts, display)

    intercepts = [
        [tangent_drop[0, 0], tangent_drop[0, 1]],
        [tangent_drop[-1, 0], tangent_drop[-1, 1]],
    ]

    # timers
    fit_preprocessing_time = time.time() - start_time
    fit_start_time = time.time()

    # start the fit
    Ndrop = np.shape(tangent_drop)[0]
    pts1 = tangent_drop[:num_points]
    pts2 = tangent_drop[-num_points:]
    pts2 = pts2[::-1]  # reverse so that CP is first

    fit_local1 = np.polyfit(pts1[:, 0], pts1[:, 1], polynomial_degree)
    fit_local2 = np.polyfit(pts2[:, 0], pts2[:, 1], polynomial_degree)

    line_local1 = np.poly1d(fit_local1)
    line_local2 = np.poly1d(fit_local2)

    if polynomial_degree > 1:
        x_local1 = np.array([min(pts1[:, 0]), max(pts1[:, 0])])
        f_local1 = line_local1(x_local1)
        f_local1_prime = line_local1.deriv(1)

        x_local2 = np.array([min(pts2[:, 0]) - 10, max(pts2[:, 0]) + 10])
        f_local2 = line_local2(x_local2)
        f_local2_prime = line_local2.deriv(1)

        tangent1 = f_local1_prime(pts1[0, 0]) * (x_local1 - pts1[0, 0]) + pts1[0, 1]
        tangent2 = f_local2_prime(pts2[0, 0]) * (x_local2 - pts2[0, 0]) + pts2[0, 1]

        m1 = f_local1_prime(pts1[0, 0])
        m2 = f_local2_prime(pts2[0, 0])
    else:
        m1 = fit_local1[0]
        m2 = fit_local2[0]

    if display:
        plt.plot(edges_pts[:, 0], edges_pts[:, 1], "o", color="pink")
        plt.plot(pts1[:, 0], pts1[:, 1], "ro")
        plt.plot(pts1[:, 0], line_local1(pts1[:, 0]), "y-")
        plt.imshow(img)
        pts1_width = abs(pts1[:, 0][-1] - pts1[:, 0][0])
        pts1_height = abs(pts1[:, 1][-1] - pts1[:, 1][0])
        plt.xlim(min(pts1[:, 0]) - pts1_width, max(pts1[:, 0]) + pts1_width)
        plt.ylim(max(pts1[:, 1]) + pts1_height, min(pts1[:, 1]) - pts1_height)
        plt.title("fitted points left side")
        plt.show()
        plt.close()

        plt.plot(edges_pts[:, 0], edges_pts[:, 1], "o", color="pink")
        plt.plot(pts2[:, 0], pts2[:, 1], "ro")
        plt.plot(pts2[:, 0], line_local2(pts2[:, 0]), "y-")
        plt.imshow(img)
        pts2_width = abs(pts2[:, 0][-1] - pts2[:, 0][0])
        pts2_height = abs(pts2[:, 1][-1] - pts2[:, 1][0])
        plt.xlim(min(pts2[:, 0]) - pts2_width, max(pts2[:, 0]) + pts2_width)
        plt.ylim(max(pts2[:, 1]) + pts2_height, min(pts2[:, 1]) - pts2_height)
        plt.title("fitted points right side")
        plt.show()
        plt.close()

    m_surf = float(CPs[1][1] - CPs[0][1]) / float(CPs[1][0] - CPs[0][0])

    if m1 > 0:
        contact_angle1 = np.pi - np.arctan((m1 - m_surf) / (1 + m1 * m_surf))
    elif m1 < 0:
        contact_angle1 = -np.arctan((m1 - m_surf) / (1 + m1 * m_surf))
    else:
        contact_angle1 = np.pi / 2

    if m2 < 0:
        contact_angle2 = np.pi + np.arctan((m2 - m_surf) / (1 + m2 * m_surf))
    elif m2 > 0:
        contact_angle2 = np.arctan((m2 - m_surf) / (1 + m2 * m_surf))
    else:
        contact_angle2 = np.pi / 2

    contact_angle1 = contact_angle1 * 180 / np.pi
    contact_angle2 = contact_angle2 * 180 / np.pi

    fit_time = time.time() - fit_start_time

    errors = polynomial_fit_errors(pts1, pts2, line_local1, line_local2, False)

    analysis_time = time.time() - start_time

    timings = {}
    timings["method specific preprocessing time"] = fit_preprocessing_time
    timings["fit time"] = fit_time
    timings["analysis time"] = analysis_time

    return [contact_angle1, contact_angle2], intercepts, errors, timings


def polynomial_fit(profile, num_points=15, polynomial_degree=2, display=False):
    """Takes in input contact angle experimental image, and outputs the fitted
    angles and intercept coordinates. For best results, preprocessing should be
    performed before calling this function.

    The first and last num_points many points are are fitted to a polynomial, and contact
    angles are calculated by taking the gradient of the polynomial at the contact
    points. The tilt of the surface is accounted for by taking the gradient between
    contact points and balancing the polynomial angle against the surface gradient.

    Coords is the coordinates of the drop outline in the form of an array or list.
    num_points is the number of points included in the polynomial fit. The default number
        of points used is 10.
    display can be set to "True" to output figures."""

    start_time = time.time()
    CPs = [profile[0], profile[-1]]

    # start the fit
    tangent_drop = profile.copy()
    Ndrop = np.shape(tangent_drop)[0]
    pts1 = tangent_drop[:num_points]
    pts2 = tangent_drop[-num_points:]
    pts2 = pts2[::-1]  # reverse so that CP is first

    fit_local1 = np.polyfit(pts1[:, 0], pts1[:, 1], polynomial_degree)
    fit_local2 = np.polyfit(pts2[:, 0], pts2[:, 1], polynomial_degree)

    line_local1 = np.poly1d(fit_local1)
    line_local2 = np.poly1d(fit_local2)

    if polynomial_degree > 1:
        # x_local1 = np.array([min(pts1[:,0])-10,max(pts1[:,0])+10])
        x_local1 = np.array([min(pts1[:, 0]), max(pts1[:, 0])])
        f_local1 = line_local1(x_local1)
        f_local1_prime = line_local1.deriv(1)

        # x_local2 = np.array([min(pts2[:,0])-10,max(pts2[:,0])+10])
        x_local2 = np.array([min(pts2[:, 0]), max(pts2[:, 0])])
        f_local2 = line_local2(x_local2)
        f_local2_prime = line_local2.deriv(1)

        tangent1 = f_local1_prime(pts1[0, 0]) * (x_local1 - pts1[0, 0]) + pts1[0, 1]
        tangent2 = f_local2_prime(pts2[0, 0]) * (x_local2 - pts2[0, 0]) + pts2[0, 1]

        tangent_lines = (
            (
                (int(x_local1[0]), int(tangent1[0])),
                (int(x_local1[1]), int(tangent1[1])),
            ),
            (
                (int(x_local2[0]), int(tangent2[0])),
                (int(x_local2[1]), int(tangent2[1])),
            ),
        )

        m1 = f_local1_prime(pts1[0, 0])
        m2 = f_local2_prime(pts2[0, 0])
    else:
        m1 = fit_local1[0]
        m2 = fit_local2[0]

        tangent_lines = (
            (
                tuple(pts1[0]),
                (int(pts1[-1][0]), int((m1 * pts1[-1][0]) + fit_local1[1])),
            ),
            (
                tuple(pts2[0]),
                (int(pts2[-1][0]), int((m2 * pts2[-1][0]) + fit_local2[1])),
            ),
        )

    if display:
        plt.plot(profile[:, 0], profile[:, 1], "o", color="pink")
        plt.plot(pts1[:, 0], pts1[:, 1], "ro")
        plt.plot(pts1[:, 0], line_local1(pts1[:, 0]), "y-")
        # plt.imshow(img)
        pts1_width = abs(pts1[:, 0][-1] - pts1[:, 0][0])
        pts1_height = abs(pts1[:, 1][-1] - pts1[:, 1][0])
        plt.xlim(min(pts1[:, 0]) - pts1_width, max(pts1[:, 0]) + pts1_width)
        plt.ylim(max(pts1[:, 1]) + pts1_height, min(pts1[:, 1]) - pts1_height)
        plt.title("fitted points left side")
        plt.show()
        plt.close()

        plt.plot(profile[:, 0], profile[:, 1], "o", color="pink")
        plt.plot(pts2[:, 0], pts2[:, 1], "ro")
        plt.plot(pts2[:, 0], line_local2(pts2[:, 0]), "y-")
        # lt.imshow(img)
        pts2_width = abs(pts2[:, 0][-1] - pts2[:, 0][0])
        pts2_height = abs(pts2[:, 1][-1] - pts2[:, 1][0])
        plt.xlim(min(pts2[:, 0]) - pts2_width, max(pts2[:, 0]) + pts2_width)
        plt.ylim(max(pts2[:, 1]) + pts2_height, min(pts2[:, 1]) - pts2_height)
        plt.title("fitted points right side")
        plt.show()
        plt.close()

    m_surf = float(CPs[1][1] - CPs[0][1]) / float(CPs[1][0] - CPs[0][0])

    if m1 > 0:
        contact_angle1 = np.pi - np.arctan((m1 - m_surf) / (1 + m1 * m_surf))
    elif m1 < 0:
        contact_angle1 = -np.arctan((m1 - m_surf) / (1 + m1 * m_surf))
    else:
        contact_angle1 = np.pi / 2

    if m2 < 0:
        contact_angle2 = np.pi + np.arctan((m2 - m_surf) / (1 + m2 * m_surf))
    elif m2 > 0:
        contact_angle2 = np.arctan((m2 - m_surf) / (1 + m2 * m_surf))
    else:
        contact_angle2 = np.pi / 2

    contact_angle1 = contact_angle1 * 180 / np.pi
    contact_angle2 = contact_angle2 * 180 / np.pi

    fit_time = time.time() - start_time

    errors = polynomial_fit_errors(pts1, pts2, line_local1, line_local2, False)

    analysis_time = time.time() - start_time

    timings = {}
    timings["fit time"] = fit_time
    timings["analysis time"] = analysis_time

    return [contact_angle1, contact_angle2], CPs, tangent_lines, errors, timings


if 0:
    from utils.misc import resource_path
    IMG_PATH = resource_path("../../RICOphobic_cropped.png")
    img = cv2.imread(IMG_PATH)

    angles, intercepts, errors, timings = polynomial_fit_img(img, display=False)

    if 1:
        print("angles: ", angles)
        print("intercetps: ", intercepts)
        print("errors: ", errors)
        print("timings: ", timings)

    print("done")



from modules.core.classes import (
    Tolerances,
    ExperimentalSetup,
    ExperimentalDrop,
    DropData,
)
from math import pi
import numpy as np
import pytest
import sys
import os

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))


project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
sys.path.insert(0, project_root)


@pytest.fixture
def test_tolerances():
    return Tolerances(
        delta_tol=1e-6,
        gradient_tol=1e-6,
        maximum_fitting_steps=100,
        objective_tol=1e-6,
        arclength_tol=1e-6,
        maximum_arclength_steps=100,
        needle_tol=1e-6,
        needle_steps=100,
    )


@pytest.fixture
def test_experimental_setup():
    setup = ExperimentalSetup()
    setup.needle_diameter_mm = 1.0

    setup.user_input_fields = {
        "drop_density": 1000.0,
        "continuous_density": 0.0,
        "pixel_mm": 100.0,
    }
    return setup


@pytest.fixture
def test_experimental_drop():
    drop = ExperimentalDrop()
    drop.drop_contour = np.array(
        [[0, 0], [1, 1], [2, 2], [3, 3], [4, 2], [5, 1], [6, 0]]
    )
    drop.contact_points = [[0, 0], [6, 0]]
    drop.pixels_to_mm = 100.0
    return drop


@pytest.fixture
def test_drop_data():
    data = DropData()
    # Initialize test parameters
    data._params = [0.0, 0.0, 1.0, 0.5, 0.0]
    data._max_s = 5.0
    data._s_points = 50
    # Manually call to generate theoretical data
    data.generate_profile_data()
    return data


def test_tolerances_initialization(test_tolerances):
    """Test that Tolerances is initialized with correct values"""
    tol = test_tolerances
    assert tol.DELTA_TOL == 1e-6
    assert tol.GRADIENT_TOL == 1e-6
    assert tol.MAXIMUM_FITTING_STEPS == 100
    assert tol.OBJECTIVE_TOL == 1e-6
    assert tol.ARCLENGTH_TOL == 1e-6
    assert tol.MAXIMUM_ARCLENGTH_STEPS == 100
    assert tol.NEEDLE_TOL == 1e-6
    assert tol.NEEDLE_STEPS == 100


def test_experimental_setup_initialization(test_experimental_setup):
    """Test that ExperimentalSetup is initialized with correct values"""
    setup = test_experimental_setup
    assert setup.needle_diameter_mm == 1.0
    assert setup.user_input_fields["drop_density"] == 1000.0
    assert setup.user_input_fields["continuous_density"] == 0.0
    assert setup.user_input_fields["pixel_mm"] == 100.0
    assert setup.drop_region is None  # Default value
    assert setup.needle_region is None  # Default value


def test_experimental_drop_initialization(test_experimental_drop):
    """Test that ExperimentalDrop is initialized with correct values"""
    drop = test_experimental_drop
    assert drop.pixels_to_mm == 100.0
    assert drop.image is None  # Default value
    assert drop.contact_angles == {}  # Default empty dict
    assert np.array_equal(
        drop.drop_contour,
        np.array([[0, 0], [1, 1], [2, 2], [3, 3], [4, 2], [5, 1], [6, 0]]),
    )
    assert drop.contact_points == [[0, 0], [6, 0]]


def test_drop_data_initialization(test_drop_data):
    """Test that DropData is initialized with correct values"""
    data = test_drop_data
    assert data.params == [0.0, 0.0, 1.0, 0.5, 0.0]
    assert data.max_s == 5.0
    assert data.s_points == 50
    assert data.parameter_dimensions == 5
    assert data.previous_guess is None
    assert data.previous_params is None


def test_drop_data_bond(test_drop_data):
    """Test that bond() returns the correct Bond number"""
    data = test_drop_data
    assert data.bond() == 0.5


def test_drop_data_apex_radius(test_drop_data):
    """Test that apex_radius() returns the correct value"""
    data = test_drop_data
    assert data.apex_radius() == 1.0


def test_drop_data_generate_profile(test_drop_data):
    """Test that generate_profile_data() creates theoretical data"""
    data = test_drop_data
    # Verify theoretical data has been generated
    assert data.theoretical_data is not None
    # Check data shape
    if data.theoretical_data is not None:
        # s_points + 1 rows
        assert data.theoretical_data.shape[0] == data.s_points + 1
        # 6 columns (x, y, phi, x_Bond, y_Bond, phi_Bond)
        assert data.theoretical_data.shape[1] == 6


def test_drop_data_params_setter():
    """Test that setting params regenerates profile"""
    # Create a new DropData instance to avoid issues with fixture
    data = DropData()
    data._max_s = 5.0
    data._s_points = 50
    data._params = [0.0, 0.0, 1.0, 0.5, 0.0]
    data.generate_profile_data()

    # Ensure theoretical data has been generated
    assert data.theoretical_data is not None
    original_shape = data.theoretical_data.shape

    # Change parameters
    new_params = [0.0, 0.0, 2.0, 0.3, 0.0]
    data.params = new_params

    # Check parameters were updated
    assert data.params == new_params

    # Check theoretical data was regenerated
    assert data.theoretical_data is not None
    assert data.theoretical_data.shape == original_shape


def test_drop_data_max_s_setter():
    """Test that setting max_s regenerates profile"""
    # Create a new DropData instance
    data = DropData()
    data._params = [0.0, 0.0, 1.0, 0.5, 0.0]
    data._s_points = 50
    data._max_s = 5.0
    data.generate_profile_data()

    # Ensure theoretical data has been generated
    assert data.theoretical_data is not None
    original_points = data.theoretical_data.shape[0]

    # Change max_s
    data.max_s = 10.0

    # Check max_s was updated
    assert data.max_s == 10.0

    # Check theoretical data was regenerated, but point count should remain the same
    assert data.theoretical_data is not None
    assert data.theoretical_data.shape[0] == original_points


def test_drop_data_s_points_setter():
    """Test that setting s_points regenerates profile"""
    # Create a new DropData instance
    data = DropData()
    data._params = [0.0, 0.0, 1.0, 0.5, 0.0]
    data._max_s = 5.0
    data._s_points = 50
    data.generate_profile_data()

    # Ensure theoretical data has been generated
    assert data.theoretical_data is not None
    original_points = data.theoretical_data.shape[0]

    # Change s_points
    data.s_points = 100

    # Check s_points was updated
    assert data.s_points == 100

    # Check theoretical data was regenerated with more points
    assert data.theoretical_data is not None
    assert data.theoretical_data.shape[0] == 101  # s_points + 1


def test_drop_data_invalid_params(test_drop_data):
    """Test that setting invalid params raises ValueError"""
    data = test_drop_data

    # Test with incorrect dimension
    with pytest.raises(ValueError):
        data.params = [0.0, 0.0, 1.0]  # Should be 5 values


def test_drop_data_invalid_max_s(test_drop_data):
    """Test that setting invalid max_s raises ValueError"""
    data = test_drop_data

    # Test with negative value
    with pytest.raises(ValueError):
        data.max_s = -1.0


def test_drop_data_invalid_s_points(test_drop_data):
    """Test that setting invalid s_points raises ValueError"""
    data = test_drop_data

    # Test with non-positive value
    with pytest.raises(ValueError):
        data.s_points = 0

    # Test with non-integer value
    with pytest.raises(ValueError):
        data.s_points = 50.5


if __name__ == "__main__":
    # Run pytest when script is executed directly
    pytest.main(["-v", __file__])



#!/usr/bin/env python
# coding=utf-8
from modules.core.classes import ExperimentalDrop, DropData

import math
import sys
import numpy as np


def initialise_parameters(experimental_drop: ExperimentalDrop, drop_data: DropData):
    omega_rotation = 0.0  # initial rotation angle (should revisit this)
    [x_apex, y_apex, radius_apex] = fit_circle(experimental_drop.drop_data)
    bond_number = calculate_Bond_number(
        experimental_drop.drop_data, x_apex, y_apex, radius_apex
    )
    drop_data.params = [x_apex, y_apex, radius_apex, bond_number, omega_rotation]
    # maybe calculate_max_s() to determine initial max_s - although current version can handle max_s being too small
    drop_data.max_s = 4.0


# fits a circle to the drop apex to calculate the (x, y) coordinate and apex radius R_0
def fit_circle(xypoints):
    lenpoints = len(xypoints)
    sumX = np.float64(0.0)
    sumY = np.float64(0.0)
    sumX2 = np.float64(0.0)
    sumY2 = np.float64(0.0)
    sumXY = np.float64(0.0)
    sumX3 = np.float64(0.0)
    sumY3 = np.float64(0.0)
    sumX2Y = np.float64(0.0)
    sumXY2 = np.float64(0.0)
    n = max(10, int(0.1 * lenpoints))  # ensure at least 10 points are used...
    if n > lenpoints:
        n = lenpoints  # if there is not enough points take all points
    for k in range(0, n):
        xk, yk = xypoints[k]
        sumX += xk
        sumY += yk
        sumX2 += xk**2
        sumY2 += yk**2
        sumXY += xk * yk
        sumX3 += xk**3
        sumY3 += yk**3
        sumX2Y += (xk**2) * yk
        sumXY2 += xk * (yk**2)
    d11 = n * sumXY - sumX * sumY
    d20 = n * sumX2 - sumX**2
    d02 = n * sumY2 - sumY**2
    d30 = n * sumX3 - sumX2 * sumX
    d03 = n * sumY3 - sumY2 * sumY
    d21 = n * sumX2Y - sumX2 * sumY
    d12 = n * sumXY2 - sumX * sumY2
    x = ((d30 + d12) * d02 - (d03 + d21) * d11) / (2 * (d20 * d02 - d11**2))
    y = ((d03 + d21) * d20 - (d30 + d12) * d11) / (2 * (d20 * d02 - d11**2))
    c = (sumX2 + sumY2 - 2 * x * sumX - 2 * y * sumY) / n
    R = math.sqrt(c + x**2 + y**2)
    return [x, y - R, R]


# calculates the initial guess for the Bond number using method
# from Neeson et al. (see Mathematica Notebook - CalculatingBond.nb)


def calculate_Bond_number(xypoints, x_apex, y_apex, radius_apex):
    r_z2 = scaled_radius_at_scaled_height(xypoints, x_apex, y_apex, radius_apex, 2)
    if r_z2 > 0:  # JB edit 26/3/15
        bond = (
            0.1756 * r_z2**2 + 0.5234 * r_z2**3 - 0.2563 * r_z2**4
        )  # interpolated from numerical data
        return bond
    r_z1 = scaled_radius_at_scaled_height(xypoints, x_apex, y_apex, radius_apex, 1)
    #    if r_z1 > 0: #JB edit 26/3/15
    #        bond = 5.819 * (r_z1 - 1) # interpolated from numerical data
    #        return bond
    # finally, if nether of these work, just use a naive guess
    return 0.15  # JB edit 26/3/2015


# calculates the radius of the pendant drop at z = height * R_0


def scaled_radius_at_scaled_height(xypoints, x_apex, y_apex, radius_apex, height):
    lenpoints = len(xypoints)
    points_to_return = 5  # number of data points to average over
    z_value = y_apex + height * radius_apex
    if xypoints[-1][1] < z_value:
        # print('ERROR: not enough data points to accurately guess the Bond number')
        # sys.exit(1)
        return -1
    index = 0
    while xypoints[index][1] < z_value:
        index += 1
    if (index < points_to_return) or ((lenpoints - index) < points_to_return):
        #        print('ERROR: not enough data points to average over')
        #        sys.exit(1)
        return -2
    sum_radius = 0.0
    for k in range(index - points_to_return, index + points_to_return):
        sum_radius += abs(xypoints[k][0] - x_apex)
    scaled_radius = sum_radius / (2 * points_to_return * radius_apex)
    return scaled_radius


# # determines the maximum required arc length, max_s
# def calculate_max_s(xypoints):
#     lenpoints = len(xypoints)
#     global s_left
#     global s_right
#     global s_max
#     s_max = 0.5 * smax
#     s_left = 0.
#     s_right = 0.
#     ptsConsider = 10
#     for i in range(lenpoints-ptsConsider,lenpoints):
#         x, y = xypoints[i]
#         minimumDistance(x, y, s_max)
#         s_max = max(s_max, s_left, s_right)



from modules.contact_angle.extract_profile import (
    extract_drop_profile,
    detect_edges,
    prepare_hydrophobic,
)
from unittest.mock import patch
from utils.enums import ThresholdSelect

from unittest.mock import patch
import matplotlib.pyplot as plt
import pytest
import numpy as np
import cv2
import os
import sys
import matplotlib

sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), "../../")))
matplotlib.use("Agg")
plt.show = lambda: None


# Fixtures ----------------------------------------------------------------


@pytest.fixture
def mock_raw_experiment():
    class RawExperiment:
        cropped_image = np.random.randint(0, 255, (100, 100), dtype=np.uint8)
        image = np.random.randint(0, 255, (100, 100), dtype=np.uint8)
        ret = None
        contour = None

    return RawExperiment()


@pytest.fixture
def mock_user_inputs():
    class UserInputs:
        threshold_method = ThresholdSelect.USER_SELECTED
        drop_region = [(10, 10), (90, 90)]
        threshold_val = 127
        original_boole = 0
        cropped_boole = 0
        threshold_boole = 0
        show_popup = 0

    return UserInputs()


@pytest.fixture
def mock_contour_data():
    return np.array(
        [
            [50, -150],
            [55, -145],
            [60, -140],
            [65, -135],
            [70, -130],
            [75, -125],
            [80, -120],
            [85, -115],
            [90, -110],
            [95, -105],
        ],
        dtype=np.float32,
    )


# Core Function Tests -----------------------------------------------------


@patch("cv2.threshold")
@patch("cv2.GaussianBlur")
@patch("cv2.findContours")
def test_manual_threshold_processing(
    mock_findContours, mock_blur, mock_threshold, mock_raw_experiment, mock_user_inputs
):
    mock_threshold.return_value = (
        127,
        np.zeros_like(mock_raw_experiment.cropped_image),
    )
    dummy_contour = np.array([[[50, 50]], [[51, 51]], [[52, 52]]], dtype=np.int32)
    mock_findContours.return_value = ([dummy_contour], None)

    extract_drop_profile(mock_raw_experiment, mock_user_inputs)

    mock_threshold.assert_called_once()
    assert mock_raw_experiment.ret == 127
    assert mock_raw_experiment.contour is not None


@patch("cv2.findContours")
@patch("cv2.threshold")
def test_otsu_auto_threshold(
    mock_threshold, mock_findContours, mock_raw_experiment, mock_user_inputs
):
    mock_user_inputs.threshold_method = ThresholdSelect.AUTOMATED
    mock_threshold.return_value = (
        127,
        np.zeros_like(mock_raw_experiment.cropped_image),
    )
    dummy_contour = np.array([[[50, 50]], [[51, 51]], [[52, 52]]], dtype=np.int32)
    mock_findContours.return_value = ([dummy_contour], None)

    extract_drop_profile(mock_raw_experiment, mock_user_inputs)

    assert 0 <= mock_raw_experiment.ret <= 255
    assert mock_threshold.call_args[0][1] == 0


@patch("cv2.findContours")
def test_contour_trimming(mock_findContours):
    test_contour = np.array([[[0, 0]], [[99, 99]], [[50, 50]]], dtype=np.int32)
    mock_findContours.return_value = ([test_contour], None)

    class DummyRawExp:
        image = np.ones((100, 100), dtype=np.uint8)

    contour, _ = detect_edges(
        np.zeros((100, 100), dtype=np.uint8), DummyRawExp(), None, 1, 127
    )

    assert not any(
        (contour[:, 0] == 0)
        | (contour[:, 0] == 99)
        | (contour[:, 1] == 0)
        | (contour[:, 1] == 99)
    )


# Clustering & Contact Point Tests ----------------------------------------


@pytest.fixture
def mock_contour_data():
    return np.array(
        [
            [40, 160],
            [45, 155],
            [50, 150],
            [55, 145],
            [60, 140],
            [65, 135],
            [70, 130],
            [75, 125],
            [80, 120],
            [85, 115],
            [90, 110],
            [95, 105],
            [100, 100],
        ],
        dtype=np.float32,
    )


@patch("cv2.findContours")
def test_invalid_threshold_handling(mock_findContours, mock_raw_experiment):
    dummy_contour = np.array([[[10, 10]], [[20, 20]], [[30, 30]]], dtype=np.int32)
    mock_findContours.return_value = ([dummy_contour], None)

    contour, _ = detect_edges(
        np.zeros((100, 100), dtype=np.uint8), mock_raw_experiment, None, 1, 300
    )
    assert contour is not None and len(contour) > 0


def test_empty_input_handling():
    with pytest.raises(IndexError):
        prepare_hydrophobic(np.empty((0, 2)))


# Edge Case Tests ---------------------------------------------------------


@patch("cv2.findContours")
def test_low_contrast_image_processing(mock_findContours):
    low_contrast_img = np.full((100, 100), 127, dtype=np.uint8)

    class DummyRawExp:
        image = low_contrast_img

    dummy_contour = np.array([[[50, 50]], [[51, 51]], [[52, 52]]], dtype=np.int32)
    mock_findContours.return_value = ([dummy_contour], None)

    contour, _ = detect_edges(low_contrast_img, DummyRawExp(), None, 1, 127)
    assert len(contour) > 0


@patch("cv2.findContours")
def test_high_noise_handling(mock_findContours):
    mock_findContours.return_value = (
        [np.random.randint(0, 100, (50, 1, 2), dtype=np.int32) for _ in range(10)],
        None,
    )

    class DummyRawExp:
        image = np.zeros((100, 100), dtype=np.uint8)

    contour, _ = detect_edges(
        np.zeros((100, 100), dtype=np.uint8), DummyRawExp(), None, 1, 127
    )
    assert len(contour) > 0


# Main Execution ----------------------------------------------------------


if __name__ == "__main__":
    pytest.main()
